/*

PGPMenu 0.93
Copyright (C) 1998 by Chris Wiegand

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version. This program is distributed in the hope that it will be useful,but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public Licensealong
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
// used for printf() and scanf()
#include <unistd.h>
// used for execlp()
#include <string.h>
// used for strcmp() and strcpy() and strncpy()
#include <stdlib.h>
// used for getenv()
#include "config.h"
// local configuration stuff...

// ----------------------------------------------------------------

// Eventually, this code area will access ~/.pgp4pinerc (the one that PAPP supports). 

int getPGPVersion()
{
	return PGP_VERSION;
}

int prefsAlwaysEncrypt()
{
	return ALWAYS_ENCRYPT;
}

int prefsAlwaysSign()
{
	return ALWAYS_SIGN;
}

int prefsAddNewKeys()
{
	return AUTO_ADD_KEYS;
}

int prefsEncryptToSelf()
{
	return ENCRYPT_TO_SELF;
}

int prefsOverrideRecipients()
{
	return OVERRIDE_RECIPIENTS;
}

char *prefsMyAddress()
{
	// Only used by PGP2 and GPG for the encryptToSelf area...
#ifdef MY_EMAIL_ADDRESS
	return MY_EMAIL_ADDRESS;
#else
	return " ";
#endif
}

// -----------------------------------------------------------------

void clearScreen()
{
	int i;
	for (i=1;i<60;i++) printf("\n");
}

void runProcessFG(char *externalBin)
{
    int i;
    int procID;

    // This function is meant to run a program foreground, not to background it like unix normally does...

    procID = fork();
    if (procID > 0) {
        // we're the parent...
        wait(&procID);
    } else {
#ifdef MY_DEBUG
    	printf("Running: %s\n",externalBin);
#endif
        execlp("/bin/sh","sh","-c",externalBin,NULL);	
    }
}

int askInteger(char *prompt,int defaultVal)
{
	char inChar[3]; // we only care about 1st char, but still...
	int returnVal;

	printf("%s",prompt);
	if (defaultVal > 0) { printf(" (Y/n) ? "); } else { printf(" (y/N) ? "); }
	scanf("%s",&inChar);
	returnVal = defaultVal;
	if (strcmp(inChar,"y") == 0 || strcmp(inChar,"Y") == 0) returnVal = 1;
	if (strcmp(inChar,"n") == 0 || strcmp(inChar,"N") == 0) returnVal = -1;
	return returnVal;
}

char *pathToBinary(int binNum)
{
	char *pathStr, *retVal;
	/* binNum table:
		encrypting, not signing = 1
		encrypting, signing = 2
		decrypting = 3
		key management = 4
	*/

#ifdef MY_PGP_PATH
	pathStr = MY_PGP_PATH;
#else
	pathStr = "";
#endif

	switch (getPGPVersion()) {
	case 5:		
		switch (binNum) {
		case 1:
			retVal = "pgpe"; break;
		case 2:
			retVal = "pgps"; break;
		case 3:
			retVal = "pgpv"; break;
		case 4:
			retVal = "pgpk"; break;
		}
		break;
	case 2:
		retVal = "pgp2"; break;
	case 1:
		retVal = "gpg"; break;			
	}
	return retVal;	
}

// ---------------------------------------------------------------------------------

void fileVerifyDecrypt(char *inFile,char *outFile)
{
	char pgpAppString[MAX_COMMAND_LINE_LENGTH];
	
	strcpy(pgpAppString,pathToBinary(3));

	switch (getPGPVersion()) {
	case 5:
		strcpy(pgpAppString,pathToBinary(3));
		if (prefsAddNewKeys() == 0) strcat(pgpAppString, " -K ");
		strcat(pgpAppString," -o ");
		strcat(pgpAppString,outFile);
		strcat(pgpAppString," ");
		strcat(pgpAppString,inFile); 
		break;
	case 2:
		if (prefsAddNewKeys() == 1) {
			// note, pgp2 doesn't support importing keys from decrypt...
			// so, we'll scan for new keys...
			strcpy(pgpAppString,pathToBinary(4));
			strcat(pgpAppString," -ka ");
			strcat(pgpAppString,inFile);
			runProcessFG(pgpAppString);
		}
		strcpy(pgpAppString,pathToBinary(3));
		strcat(pgpAppString," ");
		strcat(pgpAppString,inFile);
		strcat(pgpAppString," -o ");
		strcat(pgpAppString,outFile);
		break;
	case 1:
		if (prefsAddNewKeys() == 1) {
			// note, gpg 0.34 doesn't support importing keys from decrypt...
			// so, first we'll run gpg to scan for new keys...
			strcpy(pgpAppString,pathToBinary(4));
			strcat(pgpAppString," --import ");
			strcat(pgpAppString,inFile);
			runProcessFG(pgpAppString);
		}
		strcpy(pgpAppString,pathToBinary(3));
		strcat(pgpAppString," ");
		strcat(pgpAppString," --decrypt ");
		strcat(pgpAppString,inFile);
		strcat(pgpAppString," -output ");
		strcat(pgpAppString,outFile);
		break;
	
	}
	runProcessFG(pgpAppString);
}

// ---------------------------------------------------------------------------------

void fileSignEncrypt(char *inFileName, char *outFileName, int sign, int encrypt, char *recipient, int asciiArmor, int universalText, int encryptToSelf)
{
	char pgpAppString[MAX_COMMAND_LINE_LENGTH];
	char *tmpEmail;
	int j;

	if (encrypt > 0) {
		strcpy(pgpAppString,pathToBinary(1));
	} else {
		strcpy(pgpAppString,pathToBinary(2));
	}

	switch (getPGPVersion()) {
	case 5:
		if (encrypt > 0) {
			tmpEmail = strtok(recipient," ");
			strcat(pgpAppString," -r ");
			strcat(pgpAppString,tmpEmail);
		 	do {
				tmpEmail = strtok('\0',", ");
				if (tmpEmail) {
					strcat(pgpAppString," -r ");
					strcat(pgpAppString,tmpEmail);	
				}
			} while(tmpEmail);
			if (sign > 0) strcat(pgpAppString," -s ");
		}
		if (asciiArmor > 0) strcat(pgpAppString," -a ");
		if (universalText > 0) strcat(pgpAppString," -t ");
		if ((encryptToSelf > 0) && (encrypt > 0)) strcat(pgpAppString," --EncryptToSelf=1 ");
		strcat(pgpAppString," -o ");
                strcat(pgpAppString,outFileName);
                strcat(pgpAppString," ");
                strcat(pgpAppString,inFileName);
		break;
	case 2:
		if (asciiArmor > 0) strcat(pgpAppString," -a ");
		if (encrypt > 0) {
			strcat(pgpAppString," -e");
			if (sign > 0) strcat(pgpAppString,"s");
			strcat(pgpAppString," ");
			strcat(pgpAppString,inFileName);
			strcat(pgpAppString," ");
			if (encryptToSelf == 1) strcat(pgpAppString,prefsMyAddress());
			strcat(pgpAppString,recipient);
			strcat(pgpAppString," ");
		} else {
			if (sign > 0) {
				strcat(pgpAppString," -s ");
				strcat(pgpAppString,inFileName);
			}
		}
		runProcessFG(pgpAppString); // now, we move the <infile>.pgp to <outfile>
		strcpy(pgpAppString,"mv ");
		strcat(pgpAppString,inFileName);
		strcat(pgpAppString,".pgp ");
		strcat(pgpAppString,outFileName);
		break;
	case 1:
		if (encrypt > 0) {
			strcat(pgpAppString," -e ");
			if (encryptToSelf == 1) {
				strcat(pgpAppString," -r ");
				strcat(pgpAppString,prefsMyAddress());
			}
			tmpEmail = strtok(recipient," ");
			strcat(pgpAppString," -r ");
			strcat(pgpAppString,tmpEmail);
		 	do {
				tmpEmail = strtok('\0',", ");
				if (tmpEmail) {
					strcat(pgpAppString," -r ");
					strcat(pgpAppString,tmpEmail);	
				}
			} while(tmpEmail);
		}
		if (sign > 0) strcat(pgpAppString," -s ");
		if (asciiArmor > 0) strcat(pgpAppString, " -a ");
		if (universalText > 0) strcat(pgpAppString, " -t ");
		strcat(pgpAppString," -o ");
		strcat(pgpAppString,outFileName);
		strcat(pgpAppString, " ");
		strcat(pgpAppString,inFileName);
		break;
	}

        runProcessFG(pgpAppString);
}

// ---------------------------------------------------------------------------------

void fileVerifyDecryptMenu(char *inFile,char *outFile)
{
	printf("Decrypt/Verify file\n\n");
	fileVerifyDecrypt(inFile,outFile);
	return;
}

void fileSignEncryptMenu(char *inFile, char *outFile,char *recipients)
{
	int signFile, encryptFile, encryptToSelf;
	char input;
	char recipientList[EMAIL_ADDRESS_MAX_LENGTH];
	
	printf("Encrypt/Sign file\n\n");

	signFile = prefsAlwaysSign();
	if (signFile == 0) signFile = askInteger("Do you want to sign this file/message? ",1);
	encryptFile = prefsAlwaysEncrypt();
	if (encryptFile == 0) encryptFile = askInteger("Do you want to encrypt this file/message? ",0);

	if (encryptFile > 0) {
		if (getPGPVersion() == 5) {
			encryptToSelf = prefsEncryptToSelf();
			if (encryptToSelf == 0) encryptToSelf = askInteger("Do you want to encrypt to yourself, too? ",1);
		}
		if (strlen(recipients) < 2 || prefsOverrideRecipients() == 1) {
			printf("Please enter the recipients below.\n");
			scanf("%s",&recipientList);
			strcat(recipients,recipientList);
		}
	} else {
		strcpy(recipientList,"");
	}

	if (strlen(recipients) < 3) encryptFile = -1;
	if (signFile > 0 || encryptFile > 0) fileSignEncrypt(inFile,outFile,signFile,encryptFile,recipients,1,1,encryptToSelf);
}

void printUsage()
{
	printf("Usage: pgpmenu (options)\n");
	printf("-e: Encrypt / Sign file  <-=OR=->  -d: Decrypt / Verify file\n");
	printf("-i <file>: Incoming file\n");
	printf("-o <file>: Outgoing file (OPTIONAL)\n");
	printf("-r <recipients>: Recipients (OPTIONAL)\n");
	printf("Note: recipients MUST be last!\n");
}

int main(int argc,char **argv)
{
	char commandLine[MAX_COMMAND_LINE_LENGTH];
	char recipients[EMAIL_ADDRESS_MAX_LENGTH];
	char recipientsTmp[EMAIL_ADDRESS_MAX_LENGTH];
	int i,j,upperLimit,runEditor,myClearScreen;
	char inFile[MAX_COMMAND_LINE_LENGTH];
	char outFile[MAX_COMMAND_LINE_LENGTH];
	int outgoing = 0;
	char waitChar;

	if (argc == 1) {
		printUsage();
		return -1;
        }
	if (strcmp(argv[1],"--version") == 0) {
		printf("PGPMenu 0.91\n");
		return -1;
	}
	if (strcmp(argv[1],"--usage") == 0) {
		printUsage();
		return -1;
	}
	if (argc < 3) {
		printUsage();
		return -1;
        }

	strcpy(recipients,""); // initialize it...
	strcpy(recipientsTmp,""); // initialize it too...
	runEditor = 1; // by default, when sending, run it...
	myClearScreen = 1; // by default, clean it, pine doesn't...

#ifdef MY_DEBUG
	printf("ready (1..%d)\n",argc);
#endif

	for (i=1; i<=argc-1; i++) {
#ifdef MY_DEBUG
		printf("i (argv[%d]) = %s\n",i,argv[i]);
#endif
		// This stuff can be anywhere...
		if (strcmp(argv[i],"-e") == 0) outgoing = 1;
		if (strcmp(argv[i],"-d") == 0) outgoing = -1;
		if (strcmp(argv[i],"--noclear") == 0) myClearScreen = -1;
		if (strcmp(argv[i],"--noedit") == 0) runEditor = -1;
		if (i<argc-1) {
			// stuff that reads the NEXT argv[] goes inside here...
			if (strcmp(argv[i],"-i") == 0) strcpy(inFile,argv[i+1]);
			if (strcmp(argv[i],"-o") == 0) strcpy(outFile,argv[i+1]);
			if (strcmp(argv[i],"-r") == 0) {
				for (j=i+1;j<=argc-1;j++) {
#ifdef MY_DEBUG
					printf("Sub-parsing argv[%d] = %s for recipients...\n",j,argv[j]);
#endif
					strcat(recipientsTmp,argv[j]);
					strcat(recipientsTmp," ");
				}
				strncpy(recipients,recipientsTmp,strlen(recipientsTmp)-1);
				i=argc; // so that we end...
#ifdef MY_DEBUG
				printf("Recipients: \"%s\"\n",recipients);
#endif
			}
		}
	}

	if (strlen(inFile) < 3) {
		printf("No input file found in command line. Use -i (fileName).\n");
		return -1;
	}
	if (strlen(outFile) < 3) {
		printf("Using %s as both in and out file...\n",inFile);
		strcpy(outFile,inFile);
	}

#ifndef MY_DEBUG	
	if (myClearScreen > 0) clearScreen();
#endif

	if (outgoing == 0) {
		printf("You must specify either -e to sign/encrypt or -d to verify/decrypt.\n");
		return -1;
	}
	if (outgoing == 1) {
		// sending the file/message...
		// first, run the editor...
		if (runEditor > 0) {
			strcpy(commandLine,MY_EDITOR);
			strcat(commandLine," ");
			strcat(commandLine,inFile);
			runProcessFG(commandLine);
		}
		fileSignEncryptMenu(inFile,outFile,recipients);
	} else {
		// receiving the file/message
		fileVerifyDecryptMenu(inFile,outFile);
	}
	return 0;
}