// pgp4pine pki.c

// See common.c for the version and license.
// Copyright (C) 1998 by Chris Wiegand

#include "includes.h"

// -----------------------------------------------------------------

int pipe_pgp(const char *cmd, FILE **in, FILE **out)
{
         int pin[2], pout[2], child_pid;

         *in = *out = NULL;

         pipe(pin);
         pipe(pout);

	sprintf(debugLine,"pipe_pgp: running: %s\n",cmd); 
	documentStatus(debugLine);

         if(!(child_pid = fork()))
         {
           // We're the child.
           close(pin[1]);
           dup2(pin[0], 0);
           close(pin[0]);

           close(pout[0]);
           dup2(pout[1], 1);
           close(pout[1]);

           execl("/bin/sh", "sh", "-c", cmd, NULL);
           _exit(127);
         }

         // Only get here if we're the parent.
         close(pout[1]);
         *out = fdopen(pout[0], "r");

         close(pin[0]);
         *in = fdopen(pin[1], "w");

         return(child_pid);
}

void deallocAllKeys(struct pkiKey *killKey)
{
	struct pkiKey *nextKey2 = NULL;
	do {
		nextKey2 = killKey->nextKey;
		myFree(killKey);
		killKey = nextKey2;
		
	} while (killKey != NULL);
}

void getSecretKey()
{
	struct pkiKey *myNewKey = NULL;
	struct pkiKey *firstKey = NULL;
	struct pkiKey *oldKey = NULL;
	FILE *inFile, *outFile;
	char buf[65530], keyType[CONSOLE_IO_LINE_LENGTH], retValue[EMAIL_ADDRESS_MAX_LENGTH];
	char tmpString[CONSOLE_IO_LINE_LENGTH];
	int keyCount=0, lineType, inSec = 0;
	char tmpVal[MAX_INT_LENGTH];
	
	printf("Getting secret key(s) information");

	switch (myUserPrefs->pgpVersion) {
	case 5:
		pipe_pgp("pgpk -l",&inFile,&outFile);
		while (!feof(outFile)) {
			stillAlive();
			fgets(buf,sizeof(buf),outFile);
			strcpy(keyType,strtok(buf," "));
			sprintf(debugLine,"getSecretKey: lineIn: %s",buf);
			documentStatus(debugLine);
			lineType = 0; // default, unknown line type
			if (strcmp(keyType,"sec+") == 0 || strcmp(keyType,"sec") == 0) lineType = 1; // secret line...
			if (strcmp(keyType,"pub") == 0) lineType = 2; // public key
			if (strcmp(keyType,"sub") == 0) lineType = 3; // DH/DSS sub-key
			if (strcmp(keyType,"uid") == 0) lineType = 4; // user id
			if (strcmp(keyType,"sig") == 0) lineType = 5; // signature 
			switch (lineType) {
			case 3:
				break;
			case 5:
				break;
			case 1:
				inSec = 1;
				keyCount++;
				oldKey = myNewKey;
				myNewKey = (struct pkiKey *) myMalloc(sizeof(pkiKeyStruct));
				if (firstKey == NULL) firstKey = myNewKey;
				if (oldKey != NULL) oldKey->nextKey = myNewKey;

				// next, key size...
				strcpy(tmpString,strtok('\0'," "));
				strcpy(myNewKey->keySize,tmpString);

				// next, keyID
				strcpy(tmpString,strtok('\0'," "));
				strcpy(myNewKey->keyID,tmpString);
				
				// next, 2 dates
				strcpy(tmpString,strtok('\0'," "));
				strcpy(tmpString,strtok('\0'," "));
				// I don't use them...
				
				// next, key type...
				strcpy(tmpString,strtok('\0'," "));
				strcpy(myNewKey->keyType,tmpString);
				break;
			case 4:
				if (inSec == 1) {
					strcpy(tmpString,strtok('\0',"\n"));
					killFirstChar(tmpString);
					strcpy(myNewKey->displayName,tmpString);
					extractEmailAddress(tmpString);
					strcpy(myNewKey->emailAddress,tmpString);
				}
				break;
			case 2:
			case 0:
				// this is a public key, or other junk we don't care about, however, it may be subs or uids under
				// it, so we set myNewKey to junkKey, which we dump later...
				inSec = 0;
				break;
			}
		}
		fclose(outFile);
		break;
	case 2:
		pipe_pgp("pgp -kv ~/.pgp/secring.pgp",&inFile,&outFile);
		while (!feof(outFile)) {
			stillAlive();
			fgets(buf,sizeof(buf),outFile);
			strcpy(keyType,strtok(buf," "));
			sprintf(debugLine,"getSecretKey: lineIn: %s",buf);
			documentStatus(debugLine);
			if (strcmp(keyType,"sec") == 0) {
				keyCount++;
				oldKey = myNewKey;
				myNewKey = (struct pkiKey *) myMalloc(sizeof(pkiKeyStruct));
				if (firstKey == NULL) firstKey = myNewKey;
				if (oldKey != NULL) oldKey->nextKey = myNewKey;

				// next, key size...
				strcpy(tmpString,strtok('\0',"/"));
				strcpy(myNewKey->keySize,tmpString);

				// next, keyID
				strcpy(tmpString,strtok('\0'," "));
				strcpy(myNewKey->keyID,tmpString);
				
				// next, 1 date
				strcpy(tmpString,strtok('\0'," "));
				// I don't use it...
				
				// next, key type...
				strcpy(myNewKey->keyType,"RSA");

				// finally, user name
				strcpy(tmpString,strtok('\0',"\n"));
				strcpy(myNewKey->displayName,tmpString);
				extractEmailAddress(tmpString);
				strcpy(myNewKey->emailAddress,tmpString);
			}
		}
		fclose(outFile);
		break;
	case 1:
		pipe_pgp("gpg --list-secret-keys --with-colons --no-greeting",&inFile,&outFile);
		while (!feof(outFile)) {
			stillAlive();
			fgets(buf,sizeof(buf),outFile);
			if (strchr(buf,':') != NULL) {
				strcpy(keyType,strtok(buf,":"));
				sprintf(debugLine,"getSecretKey: lineIn: %s",buf);
				documentStatus(debugLine);
				if (strcmp(keyType,"sec") == 0) {
					keyCount++;
					oldKey = myNewKey;
					myNewKey = (struct pkiKey *) myMalloc(sizeof(pkiKeyStruct));
					if (firstKey == NULL) firstKey = myNewKey;
					if (oldKey != NULL) oldKey->nextKey = myNewKey;
	
					// in public keys, this next field is filled in
					// in secret keys, don't use it. I dunno.
					// strcpy(tmpString,strtok('\0',":"));
	
					// next, key size...
					strcpy(tmpString,strtok('\0',":"));
					strcpy(myNewKey->keySize,tmpString);

					// next, GPG key type...
					strcpy(tmpString,strtok('\0',":"));
					strcpy(myNewKey->keyType,"Unknown");
					if (strcmp(tmpString,"1") == 0) strcpy(myNewKey->keyType,"RSA");
					if (strcmp(tmpString,"16") == 0) strcpy(myNewKey->keyType,"ElGamal");
					if (strcmp(tmpString,"17") == 0) strcpy(myNewKey->keyType,"DSA");
					if (strcmp(tmpString,"20") == 0) strcpy(myNewKey->keyType,"ElGamal");
			
					// keyID
					strcpy(tmpString,strtok('\0',":"));
					strcpy(myNewKey->keyID,tmpString);
					
					// next, 1 date
					strcpy(tmpString,strtok('\0',":"));
					// I don't use it...
					
					// GPG 0.42 doesn't require this one anymore... (OLD)
					// next, two empty fields...
					// however, because they're empty, we only call strtok once...
					// strcpy(tmpString,strtok('\0',":"));
					
					// finally, user name
					strcpy(tmpString,strtok('\0',":")); // last colon...
					strcpy(myNewKey->displayName,tmpString);
					extractEmailAddress(tmpString);
					strcpy(myNewKey->emailAddress,tmpString);
				}
			}
		}
		fclose(outFile);
		break;
	}
	
	printf("\n");

	// now, let's ask which one they want to use...
	if (keyCount > 1) {
		int i, sel = 0;
		int clear = 0;
//		char ch;
		struct pkiKey *currentKey;
		currentKey = firstKey;
		
		// multiple secret keys...
		printf("\nYou have multiple secret keys. Listed below are the keys. Please select one.\n");
		for (i=1;i<=keyCount;i++) {
			printf("\n%i: %s\n\tType: %s\tBits: %s\tID: %s\n",i,currentKey->displayName,currentKey->keyType,currentKey->keySize,currentKey->keyID);
			currentKey = currentKey->nextKey;
		}
		printf("\n");
		do {
			printf("Please enter a valid selection: ");
			fgets(tmpVal, MAX_INT_LENGTH, stdin);
			sel = atoi(tmpVal);
			if (sel <= keyCount && sel >= 1) clear = 1;
		} while (clear == 0);
		currentKey = firstKey; // set back to beginning...
		for (i=1;i<sel;i++) currentKey = currentKey->nextKey; // skip to key we want
#ifdef USE_DISPLAY_NAME
		strcpy(retValue,currentKey->displayName);
#else
		strcpy(retValue,currentKey->keyID);
#endif
		strcpy(mySecretKey,retValue);
		documentStatus("Secret key chosen! Ah, the harmony in life!!!!!\n");
		deallocAllKeys(firstKey);
		printf(" ... Please wait ...\n");
	} else {
		if (keyCount == 0) {
			// no secret keys found!!!
			printf("\nWarning! It appears that you don't have a secret key defined! Please see\n your PGP/GPG documentation for assistance in making a key pair!\n");
		} else {
			// only one secret key...
			strcpy(retValue,firstKey->displayName);
			strcpy(mySecretKey,retValue);
			myFree(firstKey);
		}
	}
	documentStatus("Leaving secret area!\n");
}

// ---------------------------------------------------------------------------------

void fileVerifyDecrypt(const char inFile[MAX_COMMAND_LINE_LENGTH],const char outFile[MAX_COMMAND_LINE_LENGTH])
{
	char pgpAppString[MAX_COMMAND_LINE_LENGTH];

	switch (myUserPrefs->pgpVersion) {
	case 5:
		if (myUserPrefs->addNewKeys == 0) {
			sprintf(pgpAppString,"pgpv -K -o %s %s",outFile,inFile);
		} else {
			sprintf(pgpAppString,"pgpv -o %s %s",outFile,inFile);
		}
		break;
	case 2:
		if (myUserPrefs->addNewKeys == 1) {
			// note, pgp2 doesn't support importing keys from decrypt...
			// so, we'll scan for new keys...
			sprintf(pgpAppString,"pgp -ka %s",inFile);
			runProcessFG(pgpAppString);
		}
		sprintf(pgpAppString,"pgp %s -o %s",inFile,outFile);
		break;
	case 1:
		if (myUserPrefs->addNewKeys == 1) {
			// Unsure if this applies to gpg 0.4:
			// note, gpg 0.34 doesn't support importing keys from decrypt...
			// sadly, unless the key is the FIRST thing there, it won't import it...
			// so, it's useless to try importing, at least for now...
		}
		sprintf(pgpAppString,"gpg --no-greeting --output %s --decrypt %s",outFile,inFile);
		break;	
	}
	runProcessFG(pgpAppString);
}

// ---------------------------------------------------------------------------------

void fileSignEncrypt(const char inFileName[MAX_COMMAND_LINE_LENGTH],
		     const char outFileName[MAX_COMMAND_LINE_LENGTH],
		     const int sign, const int encrypt,
		     char recipient[EMAIL_ADDRESS_MAX_LENGTH], 
		     const int asciiArmor, const int universalText,
		     int encryptToSelf)
{
	char pgpAppString[MAX_COMMAND_LINE_LENGTH];
	char pgpOptions[MAX_COMMAND_LINE_LENGTH], recipList[EMAIL_ADDRESS_MAX_LENGTH];

	strcpy(pgpOptions," ");
	strcpy(recipList," ");
	strcat(pgpOptions,myUserPrefs->extraOptions);

	if (myUserPrefs->encryptToSelf > 0) {
		if (strlen(myUserPrefs->myAddress) > 3) {
			strcat(recipient," ");
			strcat(recipient,myUserPrefs->myAddress);
			strcat(recipient," ");
		} else {
			if (myUserPrefs->pgpVersion != 5) printf("Because you didn't specify your email address in ~/.pgp4pinerc,\nI can't auto-encrypt to you as well.\n");
			myUserPrefs->encryptToSelf = -1;
		}
	}

	switch (myUserPrefs->pgpVersion) {
	case 5:
		// get recipList...
		if (encrypt > 0) {
			parseRecipientList(recipient);
			strcpy(recipList,recipient);
 			if (sign > 0) strcat(pgpOptions," -s ");
		}
		if (sign > 0 && strlen(mySecretKey) > 3) {
 				strcat(pgpOptions," -u \"");
 				strcat(pgpOptions,mySecretKey);
 				strcat(pgpOptions,"\" ");
		}
		if (myUserPrefs->asciiArmor > 0) strcat(pgpOptions," -a ");
		if (myUserPrefs->universalText > 0) strcat(pgpOptions," -t ");

		if (encrypt > 0)
			sprintf(pgpAppString,"pgpe %s %s -o %s %s",pgpOptions,recipList,outFileName,inFileName);
		else
			sprintf(pgpAppString,"pgps %s -o %s %s",pgpOptions,outFileName,inFileName);
		break;
	case 2:
		if (myUserPrefs->asciiArmor > 0) strcat(pgpOptions," -a ");
		if (encrypt > 0) {
			strcat(pgpOptions," -e");
			if (sign > 0) strcat(pgpOptions,"s ");
			strcpy(recipList,recipient);
		} else {
			if (sign > 0) strcat(pgpOptions," -s ");
		}
		if (sign > 0 && strlen(mySecretKey) > 3) {
			strcat(pgpOptions," -u \"");
			strcat(pgpOptions,mySecretKey);
			strcat(pgpOptions,"\" ");
		}
		if (encrypt > 0)
			sprintf(pgpAppString,"pgp %s %s",pgpOptions,inFileName);
		else
			sprintf(pgpAppString,"pgp %s %s %s",pgpOptions,inFileName,recipList);
		runProcessFG(pgpAppString); // now, we move the <infile>.pgp to <outfile>
		if (myUserPrefs->asciiArmor > 0)
			sprintf(pgpAppString,"mv %s.asc %s",inFileName,outFileName);
		else
			sprintf(pgpAppString,"mv %s.pgp %s",inFileName,outFileName);
		break;
	case 1:
		if (encrypt > 0) {
			strcat(pgpOptions," -e ");
			parseRecipientList(recipient);
			strcpy(recipList,recipient);
			if (myUserPrefs->encryptToSelf == 1) {
				strcat(recipList," -r ");
				strcat(recipList,myUserPrefs->myAddress);
			}
		}
		if (sign > 0) {
			strcat(pgpOptions," -s ");
			if (strlen(mySecretKey) > 3) {
				strcat(pgpOptions," -u \"");
				strcat(pgpOptions,mySecretKey);
				strcat(pgpOptions,"\" ");
			}
		}
		if (myUserPrefs->asciiArmor > 0) strcat(pgpOptions, " -a ");
		if (myUserPrefs->universalText > 0) strcat(pgpOptions, " -t ");
		// if we're not encrypting, recipList is just a " ", so we're safe...
		sprintf(pgpAppString,"gpg --no-greeting %s %s -o %s %s",pgpOptions,recipList,outFileName,inFileName);
		break;
	}
        runProcessFG(pgpAppString);
}

