package PGPEnvelope::GPG::Keyring;

use strict;
use FileHandle;
use PGPEnvelope::GPG::Key;
1;

sub new {
  my $proto  = shift;
  my $class = ref ( $proto ) || $proto;
  my $self = {
              command        => 'gpg',
							'keys'           => [],
              keyHexIDs      => [],
							keySearchIDs   => [],
							keyInfo        => [],
							isLoaded       => 0,
							keyserver      => ''
             };
  bless ( $self, $class );
  return $self;
}

sub importKeys {
  my $self = shift;
  if ( $self->{keyserver} ) {
    system $self->{command}, '--keyserver', $self->{keyserver},
    '--recv-keys', @_;
  }
  else {
    system $self->{command}, '--recv-keys', @_;
  }
}


sub loadKeys {
  my $self = shift;
  
  print "processing keyring...\n";
  
  $self->{keySearchIDs} = [ @_ ];
  $self->{keyHexIDs}    = [];
  $self->{keyInfo}      = [];
  $self->{'keys'}         = []; 
  
  my $gpg = new FileHandle
    $self->{command} . ' --list-keys ' .
      join( ' ', @{ $self->{keySearchIDs} } ) . " | ";
  
  my $key = new PGPEnvelope::GPG::Key;
  
  my $i = 1;
  my $j = 0;
  while ( my $line = $gpg->getline() ) {
    push @{ $key->{info} }, $line;
    if ( $line eq "\n" ) {
      $key->parseInfo();
      $j = 0;
      foreach my $infoLine ( @{ $key->{info} } ) {
	if ( $j == 0 ) {
	  push @{ $self->{keyInfo} }, "$i) ", $infoLine;
	  $j = 1;
	}
	else {
	  push @{ $self->{keyInfo} }, ' ' x ( length( "$i" ) + 2 ), $infoLine;
	}
      }
      push @{ $self->{keyHexIDs} }, $key->{keyID};
      push @{ $self->{'keys'} }, $key;
      $i++;
      $key = new PGPEnvelope::GPG::Key;
    }
  }
  $gpg->close();
  $self->{isLoaded} = 1;
}



sub printKeys {
  my $self = shift;
  system "$self->{command} --list-keys " . join( ' ', @_ ) . ' > /dev/tty';
}


sub printKeysWithSignatures {
  my $self = shift;
  system "$self->{command} --list-keys " . join( ' ', @_ ) . ' > /dev/tty';
}
