package PGPEnvelope::PGP5::Keyring;

use strict;
use FileHandle;
use PGPEnvelope::PGP5::Key;
1;

sub new {
  my $proto  = shift;
  my $class = ref ( $proto ) || $proto;
  my $self = {
              command        => 'pgpk',
							'keys'           => [],
              keyHexIDs      => [],
							keySearchIDs   => [],
							keyInfo        => [],
							isLoaded       => 0,
							keyserver      => ''
             };
  bless ( $self, $class );
  return $self;
}


sub importKeys {
  my $self = shift;
  if ( $self->{keyserver} ) {
    system "$self->{command} +HTTPKeyServerHost=$self->{keyserver} -a " .
      join( ' ', @_ ) . ' > /dev/tty';
  }
  else {
    system "$self->{command} -a " . join( ' ', @_ ) . ' > /dev/tty';
  }
}


sub loadKeys {
  my $self = shift;
  
  print "processing keyring...\n";
  
  $self->{keySearchIDs} = [ @_ ];
  $self->{keyHexIDs}    = [];
  $self->{keyInfo}      = [];
  $self->{'keys'}         = []; 
  
  my $gpg = new FileHandle
    $self->{command} . ' -l ' .
      join( ' ', @{ $self->{keySearchIDs} } ) . ' | ';
  
  $gpg->getline();  # drop the first line for pgp5
  
  my $key = new PGPEnvelope::PGP5::Key;
  
  my $i = 1;
  my $j = 0;
  while ( my $line = $gpg->getline() ) {
    if ( $line =~ /^\d+ matching keys found/ ) {
      last;
    }
    push @{ $key->{info} }, $line;
    if ( $line eq "\n" ) {
      $key->parseInfo();
      $j = 0;
      foreach my $infoLine ( @{ $key->{info} } ) {
	if ( $j == 0 ) {
	  push @{ $self->{keyInfo} }, "$i) ", $infoLine;
	  $j = 1;
	}
	else {
	  push @{ $self->{keyInfo} }, ' ' x ( length( $i ) + 2 ), $infoLine;
	}
      }
      push @{ $self->{keyHexIDs} }, $key->{keyID};
      push @{ $self->{'keys'} }, $key;
      $i++;
      $key = new PGPEnvelope::PGP5::Key;
    }
  }
  $gpg->close();
  $self->{isLoaded} = 1;
}



sub printKeys {
  my $self = shift;
  system "$self->{command} -l " . join( ' ', @_ ) . ' > /dev/tty'; 
}


sub printKeysWithSignatures {
  my $self = shift;
  system "$self->{command} -ll " . join( ' ', @_ ) . ' > /dev/tty';
}
