package PGPEnvelope::PrefsFile;

use strict;
use PGPEnvelope::SetupFile;
use vars qw/ @ISA /;
@ISA = qw/ PGPEnvelope::SetupFile /;
1;

sub new {
  my $proto  = shift;
  my $class = ref ($proto) || $proto;
  my $self = new PGPEnvelope::SetupFile @_;
  
  $self->{DEFAULT} =
    [ "# $PGPEnvelope::Setup::VERSION",
      '# ',
      '# These are the programs used for encrypting/decrypting.',
      '# Currently gpg and pgp5 are supported.',
      'EncryptionProgram = gpg',
      'DecryptionProgram = gpg',
      '',
      '# These are the default settings for the menus.',
      '# Recipient settings refers to both the settings in the normal menu',
      '# and the end-recipient settings in the remailer menu.',
      '# These are boolean values.',
      'EncryptToRecipients = 0',
      'SignForRecipients   = 1',
      '',
      '# This sets whether you want all outbound encrypted e-mails to be',
      '# encrypted to a separate default key.  This is useful so you can read',
      '# your own sent mail.  Space-delimited.',
      '# e.g. EncryptToExtraKeys = 0xMyPersonalKeyHexID',
      'EncryptToExtraKeys = ',
      '',
      '# This specificies whether you want GPG-created messages to be',
      '# compatible with PGP 5.0(i).',
      'PGP50compatible = 1',
      '',
      '# This setting is what string, if any, put into the Comment',
      '# field of encrypted/signed messages',
      '# This overrides any setting in configuration files',
      'Comment = PGPEnvelope - http://www.bigfoot.com/~ftobin/resources.html',
      '',
      '# This specifies what sort of top and bottom border',
      '# you want around PGP decryption/verification messages',
      '# and information.  Note that PGP5 cannot separate the information from message',
      'ProcessedMessageTopBorder               = -----BEGIN PGPENVELOPE PROCESSED MESSAGE-----',
      'ProcessedMessageInformationTopBorder    = -----BEGIN PGPENVELOPE INFORMATION-----',
      'ProcessedMessageInformationBottomBorder = -----END PGPENVELOPE INFORMATION-----',
      '',
    ];
  
  
  bless ( $self, $class );
  return $self;
}


sub parseFile {
  my $self = shift;
  
  $self->{CONFIG}->define( 'EncryptionProgram',
			   { DEFAULT => 'gpg' } );

  $self->{CONFIG}->define( 'DecryptionProgram',
			   { DEFAULT => 'gpg' } );  

  $self->{CONFIG}->define( 'EncryptToRecipients',
			   { DEFAULT =>    0  } );

  $self->{CONFIG}->define( 'SignForRecipients',
			   { DEFAULT =>    1  } );

  $self->{CONFIG}->define( 'EncryptToExtraKeys',
			   { DEFAULT =>   ''  } );

  $self->{CONFIG}->define( 'PGP50compatible',
			   { DEFAULT =>    1  } );

  $self->{CONFIG}->define( 'Comment',
			   { DEFAULT => 'PGPEnvelope - http://www.bigfoot.com/~ftobin/resources.html' } );
  
  $self->{CONFIG}->define( 'ProcessedMessageTopBorder',
			   { DEFAULT => '-----BEGIN PGPENVELOPE PROCESSED MESSAGE-----' } );

  $self->{CONFIG}->define( 'ProcessedMessageInformationTopBorder',
			 { DEFAULT => '-----BEGIN PGPENVELOPE INFORMATION-----' } );

  $self->{CONFIG}->define( 'ProcessedMessageInformationBottomBorder',
			 { DEFAULT => '-----END PGPENVELOPE INFORMATION-----' } );
  
  $self->{CONFIG}->cfg_file( $self->filename() );
  
  return 1;
}

